#include "common.h"
#include <stddef.h>
#include <netdb.h>
#include <pwd.h>
#include <errno.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/wait.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <string.h>

int
internal_realloc(void **ptr,size_t new_size,size_t old_size){
  void *orig_ptr;
  ssize_t clr_len;

  if (!ptr)
    return -1;

  orig_ptr=*ptr;
  *ptr=realloc(*ptr, new_size);
  
  if (!(*ptr)) {
    *ptr=orig_ptr;
    return -2;
  }
  clr_len=new_size - old_size;
  if (clr_len > 0)
    memset( ((*ptr)+old_size),0,clr_len);
  return 0;
}


int
udp_setup_client(const char *hostname, int port,udp_con_t *con){
  int fd;
  int rc;
  struct hostent *serv_ent;
  
  serv_ent=gethostbyname(hostname);
  if (!serv_ent)
    return -1;

  memset(&(con->server),0,sizeof(con->server));
  (con->server).sin_family=AF_INET;
  (con->server).sin_port=htons(port);
  memcpy((char *)&(con->server.sin_addr),serv_ent->h_addr,serv_ent->h_length);

  fd=socket(AF_INET,SOCK_DGRAM,0);

  if (fd<0) 
    return -1;
  
  rc=bind(fd,(struct sockaddr *)&(con->server),sizeof(con->server));

  con->soc=fd;
  return fd;
}

int
udp_setup_server(udp_con_t *con,int port){
  int fd;
  int rc;

  fd=socket(AF_INET,SOCK_DGRAM,0);

  if (fd<0) 
    return -1;
  
  memset((char *)&(con->client),0,sizeof(con->client));
  (con->client).sin_family=AF_INET;
  (con->client).sin_addr.s_addr=htonl(INADDR_ANY);
  (con->client).sin_port=htons(port);

 
  rc=bind(fd,(struct sockaddr *)&(con->client),sizeof(con->client));

  con->soc=fd;
  con->port=port;
  return fd;
}
int
udp_send_message(const udp_con_t *con,const char *ipaddr,int port, const char *msg,size_t len){
  int rc;
  struct sockaddr_in addr;

  if ((!con) || (!ipaddr) || (!msg) )
    return -EINVAL;

  dbg_out("send [addr:%s port:%d]:\n %s\n",ipaddr,port,msg);

  addr.sin_family = AF_INET;
  addr.sin_port = htons(port);
  addr.sin_addr.s_addr = inet_addr(ipaddr);

  rc=sendto(con->soc,msg,len,0,(struct sockaddr *)&addr,sizeof(addr));

  return rc;
}

int
udp_send_peer(const udp_con_t *con,const char *msg,size_t len){
  int rc;

  if ((!con) || (!msg) )
    return -EINVAL;

  dbg_out("send peer:%s\n",msg);
  rc=sendto(con->soc,msg,len,0,(struct sockaddr *)&(con->server),sizeof(con->server));

  return rc;
}

int
udp_send_broadcast_with_addr(const udp_con_t *con,const char *bcast,const char *msg,size_t len){
  int rc;
  struct sockaddr_in addr;

  if ((!con) || (!msg) )
    return -EINVAL;

  udp_enable_broadcast(con);

  dbg_out("%s\n",msg);

  addr.sin_family = AF_INET;
  addr.sin_port = htons(con->port);
  if (!bcast)
    addr.sin_addr.s_addr = inet_addr("255.255.255.255");
  else
    addr.sin_addr.s_addr = inet_addr(bcast);

  rc=sendto(con->soc,msg,len,0,(struct sockaddr *)&addr,sizeof(addr));
  return rc;
}

int
udp_send_broadcast(const udp_con_t *con,const char *msg,size_t len){
  return udp_send_broadcast_with_addr(con,NULL,msg,len);
}

int
udp_enable_broadcast(const udp_con_t *con){
  int rc;
  int flag;

  if (!con) 
    return -EINVAL;

  flag=1;
  rc=setsockopt(con->soc, SOL_SOCKET, SO_BROADCAST, (void*)&flag, sizeof(int));
  if (rc<0) {
    dbg_out("Can not set broad cast:%s(%d)\n",strerror(errno),errno);
    return -errno;
  }
  return 0;
}
static int
internal_udp_set_buffer(const udp_con_t *con,int type,int max,int min,int *actual){
  int rc;
  int size;

  if ( (!con) || (!actual) )
    return -EINVAL;
  if ( (type != SO_RCVBUF) && (type != SO_SNDBUF) )
    return -EINVAL;

  size=max;
  rc=setsockopt(con->soc, SOL_SOCKET, type, (void*)&size, sizeof(int));
  if (rc<0) {
    size=min;
    rc=setsockopt(con->soc, SOL_SOCKET, type, (void*)&size, sizeof(int));
    if (rc<0) {
      dbg_out("Can not set broad cast:%s(%d)\n",strerror(errno),errno);
      *actual=size;
      return -errno;
    }
  }
  *actual=size;
  return 0;
}
static int
udp_set_buffer(const udp_con_t *con){
  int rc;
  int size=0;

  if (!con)
    return -EINVAL;

  rc=internal_udp_set_buffer(con,SO_RCVBUF,_MSG_BUF_SIZE,_MSG_BUF_MIN_SIZE,&size);
  if (size < _MSG_BUF_MIN_SIZE) {
    err_out("Can not set recv buf size:%s (%d)\n",
	    strerror(-rc),
	    -rc);
    return rc;
  }
  size=0;
  rc=internal_udp_set_buffer(con,SO_SNDBUF,_MSG_BUF_SIZE,_MSG_BUF_MIN_SIZE,&size);
  if (size < _MSG_BUF_MIN_SIZE) {
    err_out("Can not set send buf size:%s (%d)\n",
	    strerror(-rc),
	    -rc);
    return rc;
  }

  return 0;
}
int
udp_disable_broadcast(const udp_con_t *con){
  int rc;
  int flag;

  if (!con) 
    return -EINVAL;


  flag=0;
  rc=setsockopt(con->soc, SOL_SOCKET, SO_BROADCAST, (void*)&flag, sizeof(int));
  if (rc<0) {
    dbg_out("Can not set broad cast:%s(%d)\n",strerror(errno),errno);
    return -errno;
  }
  return 0;
}
int
udp_recv_message_full(const udp_con_t *con,char **msg,size_t *len){
  ssize_t recv_len;
  size_t addr_len;
  size_t current_len;
  size_t old_len;
  char recv_buf[_MSG_BUF_SIZE];
  char *wcp=NULL;
  int old_fd_flags;
  int rc;
  int ret_code=0;

  if ( (!con) || (!msg) || (!len) )
    return -EINVAL;

  _assert(*msg==NULL);  
  /* 
   *ノンブロッキングに設定
   */
  old_fd_flags=fcntl(con->soc,F_GETFL);
  if (old_fd_flags<0) {
    err_out("get fd flags failed %s(errno:%d)\n",strerror(errno),errno);
      return -errno;
  }

  rc=fcntl(con->soc,F_SETFL,(old_fd_flags|O_NONBLOCK));
  if (rc<0) {
    err_out("set fd flags failed: %s(errno:%d)\n",strerror(errno),errno);
      return -errno;
  }
  rc=udp_set_buffer(con);
  if (rc<0)
    goto err_out;

  rc=udp_enable_broadcast(con);
  if (rc<0)
    goto err_out;

  for(*len=0,current_len=0;;) {

    /*
     *  空読みする
     */
    addr_len=sizeof(con->server);
    recv_len=recvfrom(con->soc,recv_buf,_MSG_BUF_SIZE,MSG_PEEK,(struct sockaddr *)&(con->server),&addr_len);

    if (recv_len<=0) {
      if (errno == EAGAIN) {
	if (*len > 0) {
	  _assert(*msg!=NULL);
	  *((*msg)+(*len))='\0'; /*ヌル終端を付加*/
	}
	break;
      }
      err_out("%s(errno:%d)\n",strerror(errno),errno);
      ret_code=-errno;
      goto err_out;
    }

    dbg_out("read:%d %s\n",recv_len,recv_buf);
    /*
     *  バッファに読み込む
     */
    addr_len=sizeof(con->server);
    recv_len=recvfrom(con->soc,recv_buf,recv_len,0,(struct sockaddr *)&(con->server),&addr_len);
    if (recv_len<0) {
      err_out("%s(errno:%d)\n",strerror(errno),errno);
      ret_code=-errno;
      goto err_out;
    }

    /*
     * バッファ長更新
     */
    if (current_len < (recv_len + (*len) ) ) {
      old_len=current_len;
      while(current_len  < (recv_len + (*len) ) ) {
	current_len += _MSG_BUF_SIZE;
      }
      internal_realloc((void **)msg,current_len,old_len);
      dbg_out("Extended Buffer:%d from %d\n",current_len,old_len);
      wcp=*msg + (*len);
    }
    
    memmove(wcp,recv_buf,recv_len);  /*  内容をコピー  */
  
    /*
     *  ポインタおよびカウンタ更新
     */
    wcp += recv_len;
    *len += recv_len;
  }

err_out:
  rc=fcntl(con->soc,F_SETFL,old_fd_flags);
  if (rc<0) {
    err_out("set fd flags failed: %s(errno:%d)\n",strerror(errno),errno);
      if (*msg)
	free(*msg);
      return -errno;
  }
  return ret_code;
}

int
udp_recv_message(const udp_con_t *con,char **msg,size_t *len){
  ssize_t recv_len;
  size_t addr_len;
  char recv_buf[_MSG_BUF_SIZE];
  int old_fd_flags;
  int rc;
  int ret_code=0;

  if ( (!con) || (!msg) || (!len) )
    return -EINVAL;

  _assert(*msg==NULL);  
  /* 
   *ノンブロッキングに設定
   */
  old_fd_flags=fcntl(con->soc,F_GETFL);
  if (old_fd_flags<0) {
    err_out("get fd flags failed %s(errno:%d)\n",strerror(errno),errno);
      return -errno;
  }

  rc=fcntl(con->soc,F_SETFL,(old_fd_flags|O_NONBLOCK));
  if (rc<0) {
    err_out("set fd flags failed: %s(errno:%d)\n",strerror(errno),errno);
      return -errno;
  }

  rc=udp_set_buffer(con);
  if (rc<0)
    goto err_out;

  rc=udp_enable_broadcast(con);
  if (rc<0)
    goto err_out;

  /*
   *  空読みする
   */
  addr_len=sizeof(con->server);
  recv_len=recvfrom(con->soc,recv_buf,_MSG_BUF_SIZE,MSG_PEEK,(struct sockaddr *)&(con->server),&addr_len);

  if (recv_len<=0) {
    err_out("%s(errno:%d)\n",strerror(errno),errno);
    ret_code=-errno;
    goto err_out;
  }

  dbg_out("read:%d %s\n",recv_len,recv_buf);
  /*
   *  バッファに読み込む
   */
  addr_len=sizeof(con->server);
  recv_len=recvfrom(con->soc,recv_buf,recv_len,0,(struct sockaddr *)&(con->server),&addr_len);
  if (recv_len<0) {
    err_out("%s(errno:%d)\n",strerror(errno),errno);
    ret_code=-errno;
    goto err_out;
  }

  /*
   * バッファ獲得
   */
  
  _assert(*msg==NULL);
  *msg=malloc(recv_len);
  if (!(*msg))
    goto err_out;
  dbg_out("copy:%d %s\n",recv_len,recv_buf);
  *len=recv_len;
  memmove(*msg,recv_buf,recv_len);  /*  内容をコピー  */

err_out:
  rc=fcntl(con->soc,F_SETFL,old_fd_flags);
  if (rc<0) {
    err_out("set fd flags failed: %s(errno:%d)\n",strerror(errno),errno);
    if (*msg)
      free(*msg);
    return -errno;
  }
  return ret_code;
}

int 
udp_is_message_arrive(const udp_con_t *con,unsigned long tmout){
  fd_set rd_set;
  int rc;
  struct timeval tmout_val;

  if (!con)
    return -EINVAL;

 retry:
  FD_ZERO(&rd_set);
  FD_SET(con->soc,&rd_set);

  tmout_val.tv_sec=tmout;
  tmout_val.tv_usec=0;
  rc=select(con->soc+1,&rd_set,NULL,NULL,&tmout_val);

  if (rc<0) {
    if (errno == EINTR)
      goto retry;
    return -errno;
  }
  if (FD_ISSET(con->soc,&rd_set))
    return 1;
  else
    return 0;
}

